@@ -11,7 +11,7 @@ use tokio::{
     process::{Child, Command},
     signal::unix::{signal, SignalKind},
     sync::watch::{channel, Receiver},
-    time::{sleep, Duration},
+    time::{sleep, Duration, Instant},
 };
 use tracing::{debug, error, info};
 use tracing_subscriber::EnvFilter;
@@ -60,6 +60,30 @@ struct Ports {
     remote: u32,
 }
 
+/// Run a health check over SSH using curl. Return value is whether
+/// the health check passed or not.
+async fn run_health_check(connection: &str, remote_port: u32) -> bool {
+    let result = Command::new("ssh")
+        .args(&[
+            connection,
+            "curl",
+            "--max-time",
+            "10",
+            &format!("http://localhost:{}", remote_port),
+        ])
+        .kill_on_drop(true)
+        .output()
+        .await;
+    match result {
+        // TODO: print curl output on failure
+        Ok(status) => status.status.code().unwrap_or(1) == 0,
+        Err(err) => {
+            info!("Error running health check: {}", err);
+            false
+        }
+    }
+}
+
 fn create_child(ports: Ports, connection: &str) -> Result<Child> {
     Ok(Command::new("ssh")
         .args(&[
@@ -96,6 +120,38 @@ async fn run_child(
         };
         info!("{}: started", name);
         active_counter.inc();
+        let mut last_check = Instant::now();
+        loop {
+            // Only run the health check every 60s
+            if last_check.elapsed() >= Duration::from_secs(60) {
+                if !run_health_check(connection, ports.remote).await {
+                    if let Err(err) = child.kill().await {
+                        info!("{}: error killing ssh after health check failed: {}", name, err);
+                    }
+                    break;
+                } else {
+                    debug!("{}: passed health check", name);
+                }
+                last_check = Instant::now();
+            }
+            match child.try_wait() {
+                Ok(Some(_)) => {
+                    info!("{}: exited, restarting in 5", name);
+                    break;
+                }
+                Ok(None) => {
+                    // Still alive, just wait 5s before looping
+                    sleep(Duration::from_secs(5)).await;
+                }
+                Err(err) => {
+                    info!(
+                        "{}: exited with error, restarting in 5: {}",
+                        name, err
+                    );
+                    break;
+                }
+            }
+        }
         match child.wait().await {
             Ok(_) => {
                 info!("{}: exited, restarting in 5", name)
